# -*- coding: utf-8 -*-
"""
Created on Sun Apr 25 10:58:08 2021

@author: Maarten van Ormondt
"""

import os

import boto3
import toml
import yaml
from botocore import UNSIGNED
from botocore.client import Config

from .track_dataset import CycloneTrackDataset


class CycloneTrackDatabase:
    """
    The main Cyclone Track Database class
    Contains one or more track datasets.
    
    :param pth: Path name where bathymetry tiles will be cached.
    :type pth: string            
    """
    
    def __init__(self, path,
                 s3_bucket=None,
                 s3_key=None,
                 s3_region=None):
        self.path    = path
        self.dataset = []
        self.s3_client = None
        self.s3_bucket = s3_bucket
        self.s3_key = s3_key
        self.s3_region = s3_region

        self.read()
    
    def read(self):
        """
        Reads meta-data of all datasets in the database. 
        """
        # Check if the path exists. If not, create it.
        if not os.path.exists(self.path):
            os.makedirs(self.path)

        # Read in database
        tml_file = os.path.join(self.path, "cyclone_track_datasets.tml")
        if not os.path.exists(tml_file):
            print("Warning! Cyclone tracks database file not found: " + tml_file)
            return

        datasets = toml.load(tml_file)

        for d in datasets["dataset"]:

            name = d["name"]

            if "path" in d:
                path = d["path"]
            else:
                path = os.path.join(self.path, name)

            # Read the meta data for this dataset
            fname = os.path.join(path, "metadata.tml")

            if os.path.exists(fname):
                metadata = toml.load(fname)
                dataset_format = metadata["format"]
            else:
                print("Could not find metadata file for dataset " + name + " ! Skipping dataset.")
                continue

            if dataset_format.lower() == "ibtracs":
                dataset = CycloneTrackDataset(name, path)
            elif dataset_format.lower() == "hurdat2":
                pass
            
            self.dataset.append(dataset)

    def check_online_database(self):
        if self.s3_client is None:
            self.s3_client = boto3.client('s3', config=Config(signature_version=UNSIGNED))
        if self.s3_bucket is None:
            return
        # First download a copy of cyclone_track_datasets.tml and call it cyclone_track_datasets_s3.tml
        key = f"{self.s3_key}/cyclone_track_datasets.tml"
        filename = os.path.join(self.path, "cyclone_track_datasets_s3.tml")
        print("Updating cyclone track database ...")
        try:
            self.s3_client.download_file(Bucket=self.s3_bucket,     # assign bucket name
                                         Key=key,           # key is the file name
                                         Filename=filename) # storage file path
        except:
            # Download failed
            print(f"Failed to download {key} from {self.s3_bucket}. Database will not be updated.")
            return

        # Read bathymetry_s3.tml
        short_name_list, long_name_list = self.dataset_names()
        datasets_s3 = toml.load(filename)
        track_datasets_added = False
        added_names = []
        # Loop through s3 datasets, and check whether they exist in the local database.
        # If so, check if the metadata also exists. If not, make local folder and download the metadata.
        # Additionally, check if available_tiles.nc in s3 and not in local database, download it.
        for d in datasets_s3["dataset"]:
            # Get list of existing datasets
            s3_name = d["name"]
            if s3_name not in short_name_list:
                # Dataset not in local database
                print(f"Adding track dataset {s3_name} to local database ...")
                # Create folder and download metadata
                path = os.path.join(self.path, s3_name)
                os.makedirs(path, exist_ok=True)
                key = f"{self.s3_key}/{s3_name}/metadata.tml"
                filename = os.path.join(path, "metadata.tml")
                # Download metadata
                try:
                    self.s3_client.download_file(Bucket=self.s3_bucket, # assign bucket name
                                                Key=key,               # key is the file name
                                                Filename=filename)     # storage file path
                except Exception as e:
                    print(e)
                    print(f"Failed to download {key}. Skipping tide model.")
                    continue
                # Necessary data has been downloaded    
                track_datasets_added = True
                added_names.append(s3_name)
        # Write new local bathymetry.tml
        if track_datasets_added:
            d = {}
            d["dataset"] = []
            for name in short_name_list:
                d["dataset"].append({"name": name})
            for name in added_names:
                d["dataset"].append({"name": name})
            # Now write the new bathymetry.tml
            with open(os.path.join(self.path, "cyclone_track_datasets.tml"), "w") as tml:
                toml.dump(d, tml)            
            # Read the database again
            self.dataset = []
            self.read()
        # else:
        #     print("No new tide models were added to the local database.")

    def get_dataset(self, name):
        for dataset in self.dataset:
            if dataset.name == name:
                # Make sure the dataset is locally available
                dataset.download()
                dataset.read()
                return dataset
        return None

    def dataset_names(self):
        short_name_list = []
        long_name_list = []
        for dataset in self.dataset:
            short_name_list.append(dataset.name)
            long_name_list.append(dataset.long_name)
        return short_name_list, long_name_list

# def dict2yaml(file_name, dct, sort_keys=False):
#     yaml_string = yaml.dump(dct, sort_keys=sort_keys)    
#     file = open(file_name, "w")  
#     file.write(yaml_string)
#     file.close()

# def yaml2dict(file_name):
#     file = open(file_name,"r")
#     dct = yaml.load(file, Loader=yaml.FullLoader)
#     return dct



# import datetime
# from functools import reduce

# import geopandas as gpd
# import numpy as np
# import pandas as pd
# import shapely
# import xarray as xr

# from .tropical_cyclone_refactored import TropicalCyclone

# class CycloneTrackDatabase:
#     """
#     CycloneTrackDatabase is a class that reads in a database of tropical cyclone
#     tracks and provides methods to filter the database and return a geopandas
#     dataframe of the tracks.

#     Parameters
#     ----------
#     name : str
#         Name of the database. Currently only "ibtracs" is supported.
#     file_name : str
#         File name of the database. Currently only "ibtracs" is supported.

#     Attributes
#     ----------
#     name : str
#         Name of the database. Currently only "ibtracs" is supported.
#     ds : xarray.Dataset
#         Dataset of the database.
#     lon : numpy.ndarray
#         Longitude of the tracks.
#     lat : numpy.ndarray
#         Latitude of the tracks.
#     basin : list
#         Basin of the tracks.
#     name : list
#         Name of the tracks.
#     year : numpy.ndarray
#         Year of the tracks.
#     nstorms : int
#         Number of storms in the database.
#     ntimes : int
#         Number of time steps in the database.
#     """

#     def __init__(self, name, file_name=None):
#         self.name = name
#         self.file_name = file_name
#         self.ds = None
#         self.lon = None
#         self.lat = None
#         self.basin = None
#         self.name = None
#         self.year = None
#         self.nstorms = None
#         self.ntimes = None

#         if name == "ibtracs":
#             self._read_ibtracs(file_name)
#         else:
#             pass

#     def _read_ibtracs(self, file_name):
#         """
#         Read in IBTrACS database

#         Parameters
#         ----------
#         file_name : str
#             File name of the IBTrACS database.

#         Returns
#         -------
#         None
#         """

#         # Read in database
#         self.ds = xr.open_dataset(file_name)

#         # Convert to numpy arrays
#         self.lon = self.ds["lon"].values[:]
#         self.lat = self.ds["lat"].values[:]
#         self.basin = self.ds["basin"].values[:, 0].astype(str).tolist()
#         self.name = self.ds["name"].values[:].astype(str).tolist()
#         self.year = self.ds["season"].values[:].astype(int)
#         self.nstorms = np.shape(self.lon)[0]
#         self.ntimes = np.shape(self.lon)[1]

#     def get_track(self, index):
#         """
#         Get a single track from the database.

#         Parameters
#         ----------
#         index : int
#             Index of the track in the database.

#         Returns
#         -------
#         tc : TropicalCyclone
#             TropicalCyclone object of the track.
#         """

#         # Create a TropicalCyclone object
#         tc = TropicalCyclone(name=self.name[index])

#         gdf = gpd.GeoDataFrame()

#         # Add track
#         for it in range(self.ntimes):
#             # Check if track is finite
#             if not np.isfinite(self.lon[index, it]):
#                 break

#             # Create a shapely point
#             point = shapely.Point(self.lon[index, it], self.lat[index, it])

#             # Initialize data for geopandas dataframe
#             t = self.ds["time"].values[index, it]
#             tc_time_string = datetime.datetime.utcfromtimestamp(
#                 t.item() / 10**9
#             ).strftime("%Y%m%d %H%M%S")
#             vmax = self.ds["usa_wind"].values[index, it]
#             pc = self.ds["usa_pres"].values[index, it]
#             RMW = self.ds["usa_rmw"].values[index, it]
#             R35_NE = self.ds["usa_r34"].values[index, it, 0]
#             R35_SE = self.ds["usa_r34"].values[index, it, 1]
#             R35_SW = self.ds["usa_r34"].values[index, it, 2]
#             R35_NW = self.ds["usa_r34"].values[index, it, 3]
#             R50_NE = self.ds["usa_r50"].values[index, it, 0]
#             R50_SE = self.ds["usa_r50"].values[index, it, 1]
#             R50_SW = self.ds["usa_r50"].values[index, it, 2]
#             R50_NW = self.ds["usa_r50"].values[index, it, 3]
#             R65_NE = self.ds["usa_r64"].values[index, it, 0]
#             R65_SE = self.ds["usa_r64"].values[index, it, 1]
#             R65_SW = self.ds["usa_r64"].values[index, it, 2]
#             R65_NW = self.ds["usa_r64"].values[index, it, 3]
#             # R100_NE = self.ds["usa_r100"].values[index, it, 0]
#             # R100_SE = self.ds["usa_r100"].values[index, it, 1]
#             # R100_SW = self.ds["usa_r100"].values[index, it, 2]
#             # R100_NW = self.ds["usa_r100"].values[index, it, 3]
#             vmax = np.nan_to_num(vmax, copy=True, nan=-999.0)
#             pc = np.nan_to_num(pc, copy=True, nan=-999.0)
#             RMW = np.nan_to_num(RMW, copy=True, nan=-999.0)
#             R35_NE = np.nan_to_num(R35_NE, copy=True, nan=-999.0)
#             R35_SE = np.nan_to_num(R35_SE, copy=True, nan=-999.0)
#             R35_SW = np.nan_to_num(R35_SW, copy=True, nan=-999.0)
#             R35_NW = np.nan_to_num(R35_NW, copy=True, nan=-999.0)
#             R50_NE = np.nan_to_num(R50_NE, copy=True, nan=-999.0)
#             R50_SE = np.nan_to_num(R50_SE, copy=True, nan=-999.0)
#             R50_SW = np.nan_to_num(R50_SW, copy=True, nan=-999.0)
#             R50_NW = np.nan_to_num(R50_NW, copy=True, nan=-999.0)
#             R65_NE = np.nan_to_num(R65_NE, copy=True, nan=-999.0)
#             R65_SE = np.nan_to_num(R65_SE, copy=True, nan=-999.0)
#             R65_SW = np.nan_to_num(R65_SW, copy=True, nan=-999.0)
#             R65_NW = np.nan_to_num(R65_NW, copy=True, nan=-999.0)
#             R100_NE = -999.0
#             R100_SE = -999.0
#             R100_SW = -999.0
#             R100_NW = -999.0

#             # Create a geopandas dataframe
#             gdf_point = gpd.GeoDataFrame(
#                 {
#                     "datetime": [tc_time_string],
#                     "geometry": [point],
#                     "vmax": [vmax],
#                     "pc": [pc],
#                     "rmw": [RMW],
#                     "r35_ne": [R35_NE],
#                     "r35_se": [R35_SE],
#                     "r35_sw": [R35_SW],
#                     "r35_nw": [R35_NW],
#                     "r50_ne": [R50_NE],
#                     "r50_se": [R50_SE],
#                     "r50_sw": [R50_SW],
#                     "r50_nw": [R50_NW],
#                     "r65_ne": [R65_NE],
#                     "r65_se": [R65_SE],
#                     "r65_sw": [R65_SW],
#                     "r65_nw": [R65_NW],
#                     "r100_ne": [R100_NE],
#                     "r100_se": [R100_SE],
#                     "r100_sw": [R100_SW],
#                     "r100_nw": [R100_NW],
#                 }
#             )

#             # Append self
#             gdf = pd.concat([gdf, gdf_point])

#         # Replace -999.0 with NaN
#         gdf = gdf.replace(-999.0, np.nan)
#         gdf = gdf.reset_index(drop=True)
#         gdf = gdf.set_crs(crs=4326, inplace=True)

#         tc.track.gdf = gdf

#         return tc

#     def list_names(self, index=None):
#         if index is not None:
#             return [self.name[i] for i in index]
#         else:
#             return self.name

#     def to_gdf(self, index=None):
#         """
#         Returns a geopandas dataframe of the tracks.

#         Parameters
#         ----------
#         index : list
#             List of indices of the tracks to return. If None, all tracks are returned.

#         Returns
#         -------
#         gdf : geopandas.GeoDataFrame
#             Geopandas dataframe of the tracks.
#         """

#         # Initialize
#         geom = []
#         iok = []
#         if index is None:
#             index = range(self.nstorms)
#         description = []

#         # Loop over tracks
#         for ind in index:
#             lon = self.lon[ind, :]
#             lat = self.lat[ind, :]
#             lon = lon[np.isfinite(lon)]
#             lat = lat[np.isfinite(lat)]
#             if len(lon) > 1 and len(lat) > 1:
#                 geom.append(
#                     shapely.geometry.LineString(np.transpose(np.stack((lon, lat))))
#                 )
#                 iok.append(ind)
#                 description.append(self.name[ind] + " (" + str(self.year[ind]) + ")")

#         # Create geopandas dataframe
#         gdf = gpd.GeoDataFrame(crs=4326, geometry=geom)
#         names = [self.name[i] for i in iok]
#         years = self.year[iok]
#         gdf["name"] = names
#         gdf["year"] = years
#         gdf["description"] = description
#         gdf["database_index"] = iok
#         return gdf

#     def filter(
#         self,
#         name=None,
#         distance=None,
#         lon=None,
#         lat=None,
#         basin=None,
#         year=None,
#         year_min=None,
#         year_max=None,
#         vmax_min=None,
#         vmax_max=None,
#     ):
#         """
#         Filter the database and return the indices of the filtered tracks.

#         Parameters
#         ----------
#         name : str
#             Name of the track to filter by.
#         distance : float
#             Distance (km) of the track to the point (lon, lat).
#         lon : list
#             Longitude range of the tracks.
#         lat : list
#             Latitude range of the tracks.
#         basin : str
#             Basin of the tracks.
#         year : int
#             Year of the tracks.
#         year_min : int
#             Minimum year of the tracks.
#         year_max : int
#             Maximum year of the tracks.

#         Returns
#         -------
#         index : list
#             List of indices of the filtered tracks.
#         """

#         # Initialize year_min and year_max
#         if year:
#             year_min = year
#             year_max = year
#         else:
#             if not year_min:
#                 year_min = 0
#             if not year_max:
#                 year_max = 9999

#         if not vmax_min:
#             vmax_min = 0.0
#         if not vmax_max:
#             vmax_max = 9999.0

#         # Filter by basin
#         if basin:
#             ibasin = np.array(
#                 [
#                     i
#                     for i in range(len(self.basins))
#                     if self.basin[i].lower() == basin.lower()
#                 ]
#             )
#         else:
#             ibasin = np.arange(0, self.nstorms)

#         # Filter by year
#         if year_min and year_max:
#             iyear = np.where((self.year >= year_min) & (self.year <= year_max))[0]
#         else:
#             iyear = np.arange(0, self.nstorms)

#         # # Filter by vmax
#         # if vmax_min and vmax_max:
#         #     ivmax = np.where((self.year >= year_min) & (self.year <= year_max))[0]
#         # else:
#         #     iyear = np.arange(0, self.nstorms)

#         # Filter by name
#         if name:
#             iname = np.array(
#                 [
#                     i
#                     for i in range(len(self.name))
#                     if self.name[i].lower() == name.lower()
#                 ]
#             )
#         else:
#             iname = np.arange(0, self.nstorms)

#         # Filter by bounding box
#         if isinstance(lon, list) and isinstance(lat, list):
#             inear = np.where(
#                 (self.lon >= lon[0])
#                 & (self.lon <= lon[1])
#                 & (self.lat >= lat[0])
#                 & (self.lat <= lat[1])
#             )
#             ibbox = np.unique(inear[0])
#         else:
#             ibbox = np.arange(0, self.nstorms)

#         # Filter by distance
#         if distance:
#             # Compute distance of all tracks to point
#             d = CycloneTrackDatabase.compute_distance(lon, lat, self.lon, self.lat)
#             dmin = np.nanmin(d, axis=1)
#             idist = np.where(dmin < distance)[0]
#         else:
#             idist = np.arange(0, self.nstorms)

#         # Intersect all filters
#         index = reduce(np.intersect1d, (ibasin, iyear, iname, idist, ibbox))

#         return index

#     @staticmethod
#     def compute_distance(lon1, lat1, lon2, lat2):
#         """
#         Compute the distance between two points on a sphere.

#         Parameters
#         ----------
#         lon1 : numpy.ndarray
#             Longitude of the first point.
#         lat1 : numpy.ndarray
#             Latitude of the first point.
#         lon2 : numpy.ndarray
#             Longitude of the second point.
#         lat2 : numpy.ndarray
#             Latitude of the second point.

#         Returns
#         -------
#         d : numpy.ndarray
#             Distance between the points.
#         """

#         # Convert to radians
#         R = 6373.0
#         lon1 = lon1 * np.pi / 180
#         lat1 = lat1 * np.pi / 180
#         lon2 = lon2 * np.pi / 180
#         lat2 = lat2 * np.pi / 180
#         dlon = lon2 - lon1

#         # Wrap around
#         dlon[np.where(dlon < -np.pi)] = dlon[np.where(dlon < -np.pi)] + 2 * np.pi
#         dlon[np.where(dlon > np.pi)] = dlon[np.where(dlon > np.pi)] - 2 * np.pi
#         dlat = lat2 - lat1

#         # Compute distance
#         a = (np.sin(dlat / 2)) ** 2 + np.cos(lat1) * np.cos(lat2) * (
#             np.sin(dlon / 2)
#         ) ** 2
#         c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1.0 - a))
#         return R * c



# Load modules
import os

from cht_cyclones.tropical_cyclone import TropicalCyclone, TropicalCycloneEnsemble

# Define a track
tc                  = TropicalCyclone(name="Maria_forecast")

# Read a track
current_directory   = os.path.dirname(os.path.abspath(__file__))
file_path           = os.path.join(current_directory, 'TRK_COAMPS_CTCX_3_2017092006_15L')
tc.read_track(file_path,'trk')

# Define settings
# Uses typical Wind Enhance Scheme (WES) formulations
tc.include_rainfall     = True      # using empirical formulation to compute rainfall (IPET default)
tc.rainfall_factor      = 2.0       # factor to calibrate rainfall 2.0 means 2x as much as relationship

# Write as cyc
tc.convert_units_metric_imperial()
file_path           = os.path.join(current_directory, 'forecasted_track_Maria.cyc')
tc.write_track(file_path, 'ddb_cyc')

# create (regular) ASCII spiderweb
file_path           = os.path.join(current_directory, 'forecasted_track_Maria.spw')
tc.to_spiderweb(file_path)

# create (netcdf) spiderweb
file_path           = os.path.join(current_directory, 'forecasted_track_Maria.nc')
tc.to_spiderweb(file_path, format_type='netcdf')

# Write out as shapefile
file_path           = os.path.join(current_directory, 'forecasted_track_Maria.shp')
tc.track.to_file(file_path)

# Write out as geojson
file_path           = os.path.join(current_directory, 'forecasted_track_Maria.json')
tc.track.to_file(file_path, driver="GeoJSON")

# TC-FF ensemble members
tc2                     = TropicalCycloneEnsemble(name="Maria_forecast_ensembles", TropicalCyclone=tc)
tc2.dt                  = 3                             # 3 hourly
tc2.include_best_track  = 0                             # not included best-track
tc2.compute_ensemble(100)
tc2.to_shapefile(current_directory)


import copy
import os
from datetime import datetime

import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
from scipy.interpolate import CubicSpline, interp1d
from shapely.geometry import LineString, Point
from shapely.ops import unary_union

from .spiderweb import TropicalCycloneSpiderweb
from .utils import gdf_to_geojson_js, gdf_to_pli

# Settings
dateformat_module = "%Y%m%d %H%M%S"
knots_to_ms = float(0.51444)
nm_to_km = float(1.852)
nm_to_m = float(1.852) * 1000

geodesic = pyproj.Geod(ellps="WGS84")

class TropicalCycloneEnsemble:
    # Init
    def __init__(self, tropical_cyclone,
                 name="ensemble",
                 number_of_realizations=10,
                 compute_wind_fields=True,
                 dt=3,
                 tstart=None,
                 tend=None,
                 duration=None,
                 track_path=None,
                 spw_path=None,
                 mean_abs_cte24=19.0397,
                 sc_cte=1.3253,
                 mean_abs_ate24=26.224,
                 sc_ate=1.3432,
                 mean_abs_ve24=6.9858,
                 sc_ve=1.0000,
                 bias_ve=0.0,                    
                 format="ascii"):

        """A tropical cyclone ensemble contains a collection (list) of tropical cyclone tracks that are generated by perturbing the best track of a tropical cyclone."""

        # If name is not given, use the name of the tropical cyclone
        self.name = name

        self.compute_wind_fields = compute_wind_fields

        if track_path is None:
            track_path = os.getcwd()
        if spw_path is None:
            spw_path = os.getcwd()

        self.track_path = track_path    
        if not os.path.exists(self.track_path):
            os.makedirs(self.track_path)

        self.spw_path = spw_path    
        if not os.path.exists(self.spw_path):
            os.makedirs(self.spw_path)

        self.format = format    

        self.number_of_realizations = number_of_realizations
        self.dt = dt

        # Error statistics - defaults are based on NHC of 2018-2021
        self.mean_abs_cte24 = mean_abs_cte24 * nm_to_m # mean absolute error in cross-track error (CTE) in meter
        self.sc_cte = sc_cte  # auto-regression CTE; typically >1
        self.mean_abs_ate24 = mean_abs_ate24 * nm_to_m # mean absolute error in along-track error (ATE) in meter
        self.sc_ate = sc_ate  # auto-regression ATE; typically >1
        self.mean_abs_ve24 = mean_abs_ve24  # mean absolute error in wind error (VE) in knots
        self.sc_ve = sc_ve  # auto-regression VE = 1 = no auto-regression
        self.bias_ve = bias_ve  # bias per hour

        self.tropical_cyclone = copy.deepcopy(tropical_cyclone)

        # Set scale factor of the best track to 1.0
        self.tropical_cyclone.track.gdf.loc[:,"wind_scale_factor"] = pd.Series(np.zeros(len(self.tropical_cyclone.track.gdf)) + 1.0,
                                                                               index=self.tropical_cyclone.track.gdf.index)

        if tstart is None:
            self.tstart = datetime.strptime(
                self.tropical_cyclone.track.gdf.datetime[0], dateformat_module
            )
        else:
            self.tstart = tstart

        if tend is not None:
            # Do we really want to shorten the track?
            self.tropical_cyclone.track.shorten(tend=tend)
        else:
            if duration is not None:
                tend = self.tstart + pd.Timedelta(duration, unit="h")
                self.tropical_cyclone.track.shorten(tend=tend)

        # Make sure the metric track is computed
        self.tropical_cyclone.compute_metric_track()

        # The actual ensemble members
        self.members = []
        member = TropicalCycloneEnsembleMember()
        member.tropical_cyclone = self.tropical_cyclone
        self.members.append(member)

    def generate(self):
        """Compute the ensemble members."""

        # Create the ensemble members 
        for i in range(self.number_of_realizations):
            member = TropicalCycloneEnsembleMember()
            # member.name = self.name + "_member_" + str(i)
            # Copy the track from the best track but not the spiderweb wind field
            member.tropical_cyclone = copy.deepcopy(self.tropical_cyclone)
            # But remove the spiderweb wind field to save memory
            member.tropical_cyclone.spiderweb = TropicalCycloneSpiderweb()
            self.members.append(member)

        self.generate_tracks() # Generates the tracks and vmax values for the individual ensemble members

        if self.compute_wind_fields:
            self.compute_wind_fields() # Computes the wind fields by scaling

    def generate_tracks(self):

        # Create new track with constant time intervals
        # Resample the track to hourly values
        equidistant_best_track = copy.copy(self.tropical_cyclone.track_metric)
        equidistant_best_track.resample(self.dt)

        for i in range(0, self.number_of_realizations):

            print("Generating ensemble track " + str(i + 1) + " of " + str(self.number_of_realizations) + " ...")

            member = self.members[i]
            member.equidistant_best_track = equidistant_best_track

            if i>0:
                member.generate_random_track(self.tstart,
                                            self.mean_abs_ate24,
                                            self.sc_ate,
                                            self.mean_abs_cte24,
                                            self.sc_cte,
                                            self.mean_abs_ve24,
                                            self.sc_ve,
                                            self.bias_ve)

            fname = os.path.join(self.track_path, self.name + str(i).zfill(5) + ".cyc")
            member.tropical_cyclone.track.write(fname)

    def compute_wind_fields(self, option="scaling"):

        for i in range(self.number_of_realizations):

            print("Generating ensemble wind field " + str(i + 1) + " of " + str(self.number_of_realizations) + " ...")

            if option == "scaling":

                member = self.members[i]

                # Wind field of best track
                spw = copy.deepcopy(self.tropical_cyclone.spiderweb)

                phi = spw.ds["azimuth"].values
                r = spw.ds["range"].values

                # Loop over time, set eye coordinates, and scale wind and pressure field 
                for it in range(len(member.tropical_cyclone.track.gdf)):
                    x0 = member.tropical_cyclone.track.gdf.geometry[it].coords[0][0]
                    y0 = member.tropical_cyclone.track.gdf.geometry[it].coords[0][1]
                    spw.ds["longitude_eye"].values[it] = x0
                    spw.ds["latitude_eye"].values[it] = y0
                    facw = member.tropical_cyclone.track.gdf.wind_scale_factor[it]                    
                    spw.ds["wind_x"].values[it, :, :] = spw.ds["wind_x"].values[it, :, :] * facw
                    spw.ds["wind_y"].values[it, :, :] = spw.ds["wind_y"].values[it, :, :] * facw
                    # Scale pressure with Vatvani relation
                    pavg = 100 * member.tropical_cyclone.config["background_pressure"]
                    spw.ds["pressure"].values[it, :, :] = pavg - (pavg - spw.ds["pressure"].values[it, :, :]) * facw**2
                    # Check if there is a dataarray for precipitation
                    if "precipitation" in spw.ds:
                        spw.ds["precipitation"].values[it, :, :] = spw.ds["precipitation"].values[it, :, :] * facw

                    # Create circular grid
                    pp, rr = np.meshgrid(phi, r)
                    xx = rr * np.cos(pp * np.pi / 180)
                    yy = rr * np.sin(pp * np.pi / 180)
                    # Convert to degrees
                    spw.ds["lon"].values[it, :, :] = x0 + xx / (111320 * np.cos(y0 * np.pi / 180)) 
                    spw.ds["lon"].values[it, :, :] = y0 + yy / 111320


                if self.format == "ascii":
                    ext = ".spw"
                else:
                    ext = ".nc"
                fname = os.path.join(self.spw_path, self.name + str(i).zfill(5) + ext)
                spw.write(fname,
                          format=self.format,
                          background_pressure=1013.0,
                          include_rainfall=True,
                          merge_frac=0.5)    

    def to_gdf(self,
               option="tracks",
               varname="track_ensemble",
               filename=None,
               buffer=300000.0,
               only_forecast=True):

        """Return the ensemble as a GeoDataFrame. Option can be either tracks or outline. varname is used when the format is js."""

        # Get the GDF
        if option == "tracks":
            gdf = self.tracks_to_gdf()
        elif option == "outline" or option == "cone":
            gdf = self.outline_to_gdf(buffer, only_forecast)

        # Write to file 
        if filename is not None:
            # Get extension of filename
            ext = os.path.splitext(filename)[-1]
            # Get path of filename
            folder_path = os.path.dirname(filename)
            # Make path (if needed)
            if folder_path != "":
                os.makedirs(folder_path, exist_ok=True)
            if ext == ".shp":
                gdf.to_file(filename)
            elif ext == ".geojson":
                gdf.to_file(filename, driver="GeoJSON")
            elif ext == ".js":
                gdf_to_geojson_js(gdf, filename, varname=varname)
            elif ext == ".pli":
                gdf_to_pli(gdf, filename)

        return gdf    

    def tracks_to_gdf(self):
        """Return the ensemble as a GeoDataFrame."""
        geometries = []
        ids = []
        for nn, member in enumerate(self.members):
            coordinates = []
            for it in range(len(member.tropical_cyclone.track.gdf)):
                coordinates.append(
                    (member.tropical_cyclone.track.gdf.geometry[it].x, member.tropical_cyclone.track.gdf.geometry[it].y)
                )
            line = LineString(coordinates)
            geometries.append(line)
            ids.append(member.name)
        gdf = gpd.GeoDataFrame({"geometry": geometries, "id": ids})
        return gdf

    def outline_to_gdf(self, buffer, only_forecast=True):
        """Generate GeoDataFrame with outline polygon of all ensemble members"""
        pols = []
        for member in self.members:
            # Loop through all points to make linestring
            coordinates = []
            for it in range(len(member.tropical_cyclone.track.gdf)):
                t = datetime.strptime(member.tropical_cyclone.track.gdf.datetime[it], dateformat_module)
                if only_forecast and t < self.tstart:
                    # This is part of the best track
                    continue
                coordinates.append(
                    (member.tropical_cyclone.track.gdf.geometry[it].x, member.tropical_cyclone.track.gdf.geometry[it].y)
                )
            line = LineString(coordinates)
            line = (
                gpd.GeoDataFrame({"geometry": [line]})
                .set_crs(4326)
                .to_crs(3857)
                .iloc[0]["geometry"]
            )
            pol = line.buffer(buffer).simplify(0.1)
            pols.append(pol)
        outline_geom = unary_union(pols)
        gdf = gpd.GeoDataFrame({"geometry": [outline_geom]}).set_crs(3857).to_crs(4326)
        return gdf

    # # Make visual of all the tracks
    # def make_figures(self, folder_path):
    #     # Make path (if needed)
    #     os.makedirs(folder_path, exist_ok=True)

    #     # Get times
    #     datetime_ensemble = []
    #     for it in range(len(self.members[0].track)):
    #         datetime_ensemble.append(
    #             datetime.strptime(self.members[0].track.datetime[it], dateformat_module)
    #         )

    #     # Show wind speeds
    #     for nn, member in enumerate(self.members):
    #         plt.plot(datetime_ensemble, member.track.vmax)
    #     plt.show()

class TropicalCycloneEnsembleMember:

    def __init__(self):
        self.name = None
        self.tropical_cyclone = None

    def generate_random_track(self,
                              tstart_ensemble,
                              mean_abs_ate24,
                              sc_ate,
                              mean_abs_cte24,
                              sc_cte,
                              mean_abs_ve24,
                              sc_ve,
                              bias_ve):

        # Generate random track

        # Number of times in resampled best track
        ntpred = len(self.equidistant_best_track.gdf)

        # Get time step
        t0 = datetime.strptime(self.equidistant_best_track.gdf.datetime[0], dateformat_module)
        t1 = datetime.strptime(self.equidistant_best_track.gdf.datetime[1], dateformat_module)
        dt_seconds = (t1 - t0).total_seconds()
        dtd = dt_seconds / 86400

        np.random.seed()

        # Random error matrices from normal distribution
        arnd0 = np.random.randn(ntpred)
        crnd0 = np.random.randn(ntpred)
        vrnd0 = np.random.randn(ntpred)

        # Limit to -2 and +2 sigma
        arnd0 = np.maximum(np.minimum(arnd0, 2.0), -2.0)
        crnd0 = np.maximum(np.minimum(crnd0, 2.0), -2.0)
        vrnd0 = np.maximum(np.minimum(vrnd0, 2.0), -2.0)

        # Rest
        ate  = np.zeros((ntpred))
        cte  = np.zeros((ntpred))
        ve   = np.zeros((ntpred))
        x    = np.zeros((ntpred))
        y    = np.zeros((ntpred))
        vmax = np.zeros((ntpred))
        wind_scale = np.zeros((ntpred)) + 1.0

        # Correct for time
        tfac = np.sqrt(dtd)

        # go from mean absolute error to standard deviation
        sigma_ate = mean_abs_ate24 / np.sqrt(2 / np.pi)
        sigma_cte = mean_abs_cte24 / np.sqrt(2 / np.pi)

        # Check unit of intensity
        if self.equidistant_best_track.unit_intensity == "knots":
            sigma_ve = mean_abs_ve24 / np.sqrt(2 / np.pi)
        else:
            sigma_ve = mean_abs_ve24 * knots_to_ms / np.sqrt(2 / np.pi)

        # Loop over time
        forecast_timestep = 0.0

        # Find index of last time smaller or equal to tstart_ensemble
        it_start = 0
        for it in range(0, len(self.equidistant_best_track.gdf)):
            t = datetime.strptime(self.equidistant_best_track.gdf.datetime[it], dateformat_module)
            if t <= tstart_ensemble:
                it_start = it
                break

        # Coordinates of the first point
        x00 = self.equidistant_best_track.gdf.geometry[it_start].x
        y00 = self.equidistant_best_track.gdf.geometry[it_start].y

        # Initial errors 
        ate_12 = 0.0
        cte_12 = 0.0
        ve_12  = 0.0    

        # Loop through times in equidistant best track gdf
        for it in range(ntpred):

            t = datetime.strptime(self.equidistant_best_track.gdf.datetime[it], dateformat_module)
            x0 = self.equidistant_best_track.gdf.geometry[it].x
            y0 = self.equidistant_best_track.gdf.geometry[it].y
            vmax0 = self.equidistant_best_track.gdf.vmax[it]       

            x[it] = x0
            y[it] = y0
            vmax[it] = vmax0

            forecast_timestep = forecast_timestep + dtd * 24

            # If we want the variability
            if t > tstart_ensemble:

                # standard deviation scales with tfac
                arnd = tfac * sigma_ate * arnd0[it]
                crnd = tfac * sigma_cte * crnd0[it]
                vrnd = tfac * sigma_ve * vrnd0[it]

                # Limit to -2 and +2 sigma
                arnd = np.maximum(
                    np.minimum(arnd, 2 * tfac * sigma_ate), -2 * tfac * sigma_ate
                )
                crnd = np.maximum(
                    np.minimum(crnd, 2 * tfac * sigma_cte), -2 * tfac * sigma_cte
                )
                vrnd = np.maximum(
                    np.minimum(vrnd, 2 * tfac * sigma_ve), -2 * tfac * sigma_ve
                )

                # Compute new track errors
                at1 = sc_ate**dtd
                ate[it] = at1 * ate_12 + arnd
                ct1 = sc_cte**dtd
                cte[it] = ct1 * cte_12 + crnd

                # Compute track heading based on latitude / longitude of two points
                fwd_azimuth, back_azimuth, distance = geodesic.inv(x00, y00, x0, y0)

                # Along track error shift
                xa, ya, backaz = geodesic.fwd(
                    x0,
                    y0,
                    fwd_azimuth,
                    ate[it],
                    radians=False,
                )

                # Cross track error shift
                xc, yc, backaz = geodesic.fwd(
                    xa, ya, fwd_azimuth + 90.0, cte[it], radians=False
                )

                x[it] = xc
                y[it] = yc

                # Do wind speed too
                ve1 = sc_ve**dtd
                ve[it] = ve1 * ve_12 + vrnd + bias_ve * forecast_timestep
                # Let's make sure that the wind speed does not become negative
                if vmax0 + ve[it] < 5.0:
                    ve[it] = max(5.0 - vmax0, 0.0)
                vmax[it] = vmax0 + ve[it]
                wind_scale[it] = vmax[it] / vmax0

                # Save errors from last iteration
                ate_12 = ate[it]
                cte_12 = cte[it]
                ve_12  = ve[it]

        # And now interpolate x, y, vmax and wind_scale on the equidistant best track, to the original time steps

        # First make time stamp arrays for equidistant and original best track
        t0 = []
        for date in self.equidistant_best_track.gdf["datetime"]:
            t0.append(datetime.strptime(date, dateformat_module))
        t1 = []
        for date in self.tropical_cyclone.track.gdf["datetime"]:
            t1.append(datetime.strptime(date, dateformat_module))
        t0 = [date.timestamp() for date in t0]
        t1 = [date.timestamp() for date in t1]

        # Now interpolate

        # Use CubicSpline for x and y
        f = CubicSpline(t0, x)
        x1 = f(t1)
        f = CubicSpline(t0, y)
        y1 = f(t1)

        # Linear interpolation for vmax and wind_scale
        f = interp1d(t0, vmax)
        v1 = f(t1) / knots_to_ms
        f = interp1d(t0, wind_scale)
        ws1 = f(t1)

        # Also update central pressure 
        pavg = self.tropical_cyclone.config["background_pressure"]
        pc1 = np.zeros(len(t1))
        for it in range(len(t1)):
            pc1[it] = pavg - (pavg - self.tropical_cyclone.track.gdf.pc[it]) * ws1[it]**2

        # Loop over time for geometry
        points = []
        for it in range(len(t1)):
            points.append(Point(x1[it], y1[it]))
            self.tropical_cyclone.track.gdf.loc[it, "geometry"] = Point(x1[it], y1[it])

        # And set Vmax and wind scale factor
        self.tropical_cyclone.track.gdf.loc[:,"vmax"] = pd.Series(v1, index=self.tropical_cyclone.track.gdf.index)
        self.tropical_cyclone.track.gdf.loc[:,"pc"] = pd.Series(pc1, index=self.tropical_cyclone.track.gdf.index)
        self.tropical_cyclone.track.gdf.loc[:,"wind_scale_factor"] = pd.Series(ws1, index=self.tropical_cyclone.track.gdf.index)

        # Recompute metric values
        self.tropical_cyclone.compute_metric_track()
